from .abi_logic import Abi
from .events import Event, anonymous, indexed
from .inputs import Struct
from .state_mutability import payable, view, pure
from .types import *


def to_abi(abi_cls: Abi) -> list:
    return abi_cls.to_abi()



from typing import get_type_hints, TypeVar

# https://docs.soliditylang.org/en/v0.8.13/abi-spec.html#types
int24 = TypeVar("int24")
...
int256 = TypeVar("int256")

uint8 = TypeVar("uint8")
uint16 = TypeVar("uint16")
uint24 = TypeVar("uint24")
uint32 = TypeVar("uint32")
uint40 = TypeVar("uint40")
uint48 = TypeVar("uint48")
uint56 = TypeVar("uint56")
uint64 = TypeVar("uint64")
uint72 = TypeVar("uint72")
uint80 = TypeVar("uint80")
uint88 = TypeVar("uint88")
uint96 = TypeVar("uint96")
uint104 = TypeVar("uint104")
uint112 = TypeVar("uint112")
uint120 = TypeVar("uint120")
uint128 = TypeVar("uint128")
uint136 = TypeVar("uint136")
uint144 = TypeVar("uint144")
uint152 = TypeVar("uint152")
uint160 = TypeVar("uint160")
uint168 = TypeVar("uint168")
uint176 = TypeVar("uint176")
uint184 = TypeVar("uint184")
uint192 = TypeVar("uint192")
uint200 = TypeVar("uint200")
uint208 = TypeVar("uint208")
uint216 = TypeVar("uint216")
uint224 = TypeVar("uint224")
...
uint256 = TypeVar("uint256")
uint = uint256

address = TypeVar("address")
string = TypeVar("string")
bool = TypeVar("bool")



import inspect

from typing import TypeVar
from types import GenericAlias


class Struct:
    @classmethod
    def to_components(cls):
        return [
            {"name": name, **parse_input(type_)}
            for name, type_ in cls.__annotations__.items()
        ]


def _parse_struct(type_):
    return {"type": "tuple", "components": type_.to_components()}


def _parse_tuple(type_):
    return {
        "type": "tuple",
        "components": [{"name": "", **parse_input(a)} for a in type_.__args__],
    }


def _parse_list(type_):
    internal_type = type_.__args__[0]
    if inspect.isclass(internal_type) and issubclass(internal_type, Struct):
        return {"type": "tuple[]", "components": internal_type.to_components()}

    return {"type": str(internal_type).lstrip("~") + "[]", "components": []}


def _parse_type(type_):
    return {"type": str(type_).lstrip("~"), "components": []}


def _parse_python_type(type_):
    type_name = None

    if type_ == bool:
        type_name = "bool"
    elif type_ == str:
        type_name = "string"

    if type_name:
        return {"type": type_name, "components": []}
    raise NotImplementedError(f"Couldn't parse {type_}")


def parse_input(type_):
    if type(type_) == GenericAlias:
        if type_.__origin__ == list and len(type_.__args__) == 1:
            return _parse_list(type_)
        elif type_.__origin__ == tuple:
            return _parse_tuple(type_)
    elif type(type_) == TypeVar:
        return _parse_type(type_)
    elif issubclass(type_, Struct):
        return _parse_struct(type_)
    else:
        return _parse_python_type(type_)



import inspect
from typing import get_type_hints

from .events import Event
from .inputs import Struct, parse_input


class Abi:
    @classmethod
    @property
    def events(cls) -> list[Event]:
        events_list = []
        for item_name in set(dir(cls)) - set(dir(Abi)):
            item = getattr(cls, item_name)
            if inspect.isclass(item) and issubclass(item, Event):
                events_list.append(item)

        return events_list

    @classmethod
    @property
    def functions(cls) -> list[callable]:
        functions_list = []
        for item_name in set(dir(cls)) - set(dir(Abi)):
            item = getattr(cls, item_name)

            if (
                item_name != "__init__"  # constructor
                and item_name != "__annotations__"  # constants
                and not (
                    inspect.isclass(item) and issubclass(item, (Event, Struct))
                )  # events / structs
            ):
                functions_list.append(item)
        return functions_list

    # https://docs.soliditylang.org/en/v0.8.13/abi-spec.html#json
    @classmethod
    def to_abi(cls):
        abi = []

        for name, type_ in cls.__annotations__.items():
            abi.append(
                {
                    "type": "function",
                    "name": name,
                    "inputs": [],
                    "outputs": [{"name": "", **parse_input(type_)}],
                    "stateMutability": "view",
                }
            )

        for event in cls.events:
            abi.append(event.to_abi_item())

        for function in cls.functions:
            inputs = []
            outputs = []
            for name, type_ in get_type_hints(function).items():
                if name == "return":
                    outputs = [{"name": "", **parse_input(type_)}]
                else:
                    inputs.append({"name": name, **parse_input(type_)})

            abi.append(
                {
                    "type": "function",
                    "name": function.__name__,
                    "inputs": inputs,
                    "outputs": outputs,
                    "stateMutability": getattr(
                        function, "_state_mutability", "nonpayable"
                    ),
                }
            )

        return abi



from ..types import uint256, address
from ..state_mutability import payable
from ..events import Event, indexed
from .erc20 import ERC20


class WETH9(ERC20):
    @payable
    def deposit():
        ...

    def withdraw(wad: uint256):
        ...

    class Deposit(Event):
        dst: indexed(address)
        wad: uint256

    class Withdrawal(Event):
        src: indexed(address)
        wad: uint256



from .inputs import parse_input


def anonymous(cls):
    cls._anonymous = True
    return cls


def indexed(typeVar):
    typeVar._indexed = True
    return typeVar


class Event:
    @classmethod
    @property
    def signature(cls):
        inputs = [parse_input(type_)["type"] for type_ in cls.__annotations__.values()]
        return f"{cls.__name__}({','.join(inputs)})"

    @classmethod
    @property
    def selector(cls):
        raise NotImplementedError
        # return keccak(cls.signature)

    @classmethod
    def to_abi_item(cls):
        inputs = []
        for name, type_ in cls.__annotations__.items():
            inputs.append(
                {
                    "name": name,
                    "indexed": getattr(type_, "_indexed", False),
                    **parse_input(type_),
                }
            )

        return {
            "type": "event",
            "name": cls.__name__,
            "inputs": inputs,
            "anonymous": getattr(cls, "_anonymous", False),
        }



from ..abi_logic import Abi
from ..state_mutability import view
from ..types import string, uint8, uint256, address, bool
from ..events import Event, indexed


# https://eips.ethereum.org/EIPS/eip-20


class ERC20(Abi):
    # Methods
    name: string
    symbol: string
    decimals: uint8
    totalSupply: uint256

    @view
    def balanceOf(_owner: address) -> uint256:
        ...

    def transfer(_to: address, _value: uint256) -> bool:
        ...

    def transferFrom(_from: address, _to: address, _value: uint256) -> bool:
        ...

    def approve(_spender: address, _value: uint256) -> bool:
        ...

    @view
    def allowance(_owner: address, _spender: address) -> uint256:
        ...

    # Events

    class Transfer(Event):
        _from: indexed(address)
        _to: indexed(address)
        _value: uint256

    class Approval(Event):
        _owner: indexed(address)
        _spender: indexed(address)
        _value: uint256



def state_mutability_decorator(state_mutability):
    def wrapper(fn):
        _state_mutability = getattr(fn, "_state_mutability", None)
        if _state_mutability is not None:
            raise RuntimeWarning(
                f"Couldn't add @{state_mutability} for @{_state_mutability} function"
            )
        fn._state_mutability = state_mutability
        return fn

    return wrapper


pure = state_mutability_decorator("pure")
view = state_mutability_decorator("view")
payable = state_mutability_decorator("payable")



from abinator import (
    Abi,
    Event,
    Struct,
    indexed,
    payable,
    view,
    int24,
    uint24,
    uint96,
    uint128,
    uint256,
    address,
)


# part of https://github.com/Uniswap/v3-periphery/blob/main/contracts/interfaces/INonfungiblePositionManager.sol


class NonfungiblePositionManager(Abi):
    class IncreaseLiquidity(Event):
        tokenId: indexed(uint256)
        liquidity: uint128
        amount0: uint256
        amount1: uint256

    class DecreaseLiquidity(Event):
        tokenId: indexed(uint256)
        liquidity: uint128
        amount0: uint256
        amount1: uint256

    class Collect(Event):
        tokenId: indexed(uint256)
        recipient: address
        amount0: uint256
        amount1: uint256

    @view
    def positions(
        tokenId: uint256,
    ) -> tuple[
        uint96,
        address,
        address,
        address,
        uint24,
        int24,
        int24,
        uint128,
        uint256,
        uint256,
        uint128,
        uint128,
    ]:
        ...

    class MintParams(Struct):
        token0: address
        token1: address
        fee: uint24
        tickLower: int24
        tickUpper: int24
        amount0Desired: uint256
        amount1Desired: uint256
        amount0Min: uint256
        amount1Min: uint256
        recipient: address
        deadline: uint256

    @payable
    def mint(params: MintParams) -> tuple[uint256, uint128, uint256, uint256]:
        ...

    class IncreaseLiquidityParams(Struct):
        tokenId: uint256
        amount0Desired: uint256
        amount1Desired: uint256
        amount0Min: uint256
        amount1Min: uint256
        deadline: uint256

    @payable
    def increaseLiquidity(
        params: IncreaseLiquidityParams,
    ) -> tuple[uint128, uint256, uint256]:
        ...


NonfungiblePositionManager.to_abi()



from abinator import Abi, Struct, view, address

from abinator.common_abis import ERC20


class Factory(Abi):
    class Info(Struct):
        wallet: address
        defii: address
        hasAllocation: bool
        incentiveVault: address

    @view
    def getAllInfos() -> list[Info]:
        ...

    @view
    def getAllWallets() -> list[address]:
        ...


# print(Factory.functions)
# print(Factory.to_abi())


print(ERC20.Approval.to_abi_item()["inputs"])
